//
// Image.cpp: implementation for the Image class.
//
// portions adopted from RGBPixmap class on page 838 in 
//  "Computer Graphics using Open GL" by F.S. Hill,Jr
//
//
//      AUTHOR: Russell Lowke
//
//		  Date: Jan 20th 2005
//


#include "tga.h"
#include "Image.h"
#include "Matrix.h"
#include <iostream>
#include <exception>
#include <stdio.h>
#include <memory.h>
#include <cstdlib>
#include <cmath>
#include <math.h>
#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>

using namespace std;

// default constructor
Image::Image() {							
	m_width = m_height = m_xLoc = m_yLoc = 0;
	m_pixels = NULL;
}

// constructor
Image::Image(int width, int height) {		
	m_width   = width;
	m_height  = height;
	m_pixels  = new TGAPixel[m_width * m_height];	// create pixel heap
	m_xLoc    = 0;
	m_yLoc    = 0;
}

Image::Image(const Image& m) {						// copy constructor			
	*this = m;
}

Image::~Image() {									// destructor
	if (m_pixels != NULL)
		delete [] m_pixels;
	m_width = m_height = 0;
}

Image& Image::operator=(const Image& other) {      // used by copy constructor
	if (this == &other) { return *this;	}

	m_width   = other.m_width;
	m_height  = other.m_height;
	m_xLoc    = other.m_xLoc;
	m_yLoc	  = other.m_yLoc;

	TGAPixel* pixels = new TGAPixel[m_width * m_height];  // create pixel heap
	for (int x = 0; x < m_width; ++x)
		for(int y = 0; y < m_height; ++y)
			setAt(x, y, other(x, y));			   // copy pixel
	
	if (m_pixels != NULL)						   // m_pixels = NULL in default constructor
		delete [] m_pixels;						   // delete old pixels
	m_pixels = pixels;							   // assign pixel heap
	
	return *this;
}

void Image::fill(TGAPixel& color) {
	for (int x = 0; x < m_width; ++x)
		for (int y = 0; y < m_height; ++y)
			 m_pixels[m_width * y + x] = color;
}

void Image::draw() {
	// Note: glRasterPos2i has a lower-left origin, see setLoc()
	glRasterPos2i(m_xLoc, m_yLoc);
	glDrawPixels (m_width, m_height, GL_BGRA_EXT, GL_UNSIGNED_BYTE, m_pixels);
}
void Image::drawAt(int x, int y) {
	glRasterPos2i(m_xLoc = x, m_yLoc = y);
	glDrawPixels (m_width, m_height, GL_BGRA_EXT, GL_UNSIGNED_BYTE, m_pixels);
}
bool Image::clicked(int mouseX, int mouseY) {

	// is mouse within this image?
	if (mouseX >= m_xLoc && 
		mouseX <= m_xLoc + m_width &&
		mouseY >= m_yLoc && 
		mouseY <= m_yLoc + m_height)					// click detected
		return true;
	else
		return false;
}
void Image::resize(int nWidth, int nHeight) {

	TGAPixel* pixels  = new TGAPixel[nWidth * nHeight];	// create pixel heap
	double xMod = m_width  / (double) nWidth;
	double yMod = m_height / (double) nHeight;
	int	   tX, tY;

	for(int x = 0; x < nWidth; ++x) {
		for (int y = 0; y < nHeight; ++y) {
			tX = (int) (x * xMod);
			tY = (int) (y * yMod);
			if (tX > m_width)  { tX = m_width; }
			if (tY > m_height) { tY = m_height; }
			pixels[nWidth * y + x] = m_pixels[m_width * tY + tX]; 
		}
	}
	if (m_pixels != NULL)							// m_pixels = NULL in default constructor
		delete[] m_pixels;							// delete old pixels
	m_pixels = pixels;								// assign pixel heap
	m_width  = nWidth;
	m_height = nHeight;
}
//
// Loads a targa image into memory from a file.
//
void Image::loadFromFile(const char *filename, bool alpha) {
	m_xLoc = m_yLoc = 0;
	if (m_pixels)									// delete old heap
		delete[] m_pixels;

    FILE *fp = fopen (filename, "rb");

    if (fp == 0)
		cerr << "The file " << filename << " does not exist" << endl;

    fread(&m_header, sizeof(m_header), 1, fp);

	m_width  = m_header.Width;
	m_height = m_header.Height;

	m_pixels = new TGAPixel[m_width * m_height];

	if (m_header.ImagePixelSize == 32)
		fread(m_pixels, sizeof(TGAPixel), m_header.Width * m_header.Height, fp);
	else if (m_header.ImagePixelSize == 24) {
		for (int i = 0; i < m_width * m_height; ++i) {
			fread(&m_pixels[i].b, sizeof(char), 1, fp);
			fread(&m_pixels[i].g, sizeof(char), 1, fp);
			fread(&m_pixels[i].r, sizeof(char), 1, fp);
			if (alpha &&
				m_pixels[i].b == 255 && 
				m_pixels[i].g == 255 && 
				m_pixels[i].r == 255) {
				m_pixels[i].alpha = 0;
			} else
				m_pixels[i].alpha = 255;
		}
	}
	fclose(fp);
}
//
// Saves image to a targa file.
//
void Image::saveToFile(const char *filename) {

	FILE *fp = fopen(filename, "wb");

    if (fp == 0) throw exception("Image::saveToFile(): failed to save file.");

	fwrite(&m_header, sizeof(m_header), 1, fp);
	fwrite(m_pixels,  sizeof(TGAPixel), m_width * m_height, fp);

    fclose(fp);
}